

/* gencode.c */


#include "main.h"


/* The global program counter. */
long pc = 0L;


/* The list of resolved and unresolved labels. */
static Label * labelList = NIL;

/* The list of intermediate instructions. */
static Instruction * instructionList = NIL;



static void reverseInstructionList(void);
static void emitInstruction(FILE * fileptr, Instruction * instruction);
static void emit(FILE * fileptr, int opcode);
static long revbits(long val);
static long evalExpr(Expr * expr, int line);



/* Try to find the label with the given name.
   Returns NIL if it does not exist. */
Label * findLabel(char * name)
{
  Label * label;
  for(label = labelList; label != NIL;
      label = label->next)
    if(strcmp(label->name, name) == 0)
      return (label);
  return (NIL);
}


/* Add a label with the given name to the label list.
   There must not be another label with the same name already.
   Returns a pointer to the new label. */
Label * addLabel(char * name)
{
  Label * label = (Label *)allocate(sizeof(Label));
  label->name = duplicateString(name);
  label->resolved = FALSE;
  label->value = NIL;
  label->next = labelList;
  labelList = label;
  return (label);
}


/* Resolve a label to the given value. */
void resolveLabel(Label * label, Expr * value)
{
  if(label->resolved)
    error("internal error; label \"%s\" already resolved",
	  label);
  label->resolved = TRUE;
  label->value = value;

/*  if(value->type == INTEGER_EXPR)
    printf("resolving label \"%s\" to %lx\n",
	   label->name, value->value.integer); */
}


/* Generate an intermediate instruction. */
void gentemp(Opcode * opcode, Argument * args)
{
  Instruction * instruction;

  instruction = (Instruction *)allocate(sizeof(Instruction));
  instruction->opcode = opcode;
  instruction->lineNumber = lineNumber;
  instruction->args = args;
  instruction->next = instructionList;
  instructionList = instruction;

  pc += opcodeLength(opcode);
}


/* Generate actual opcode-bytes from the temporary instructions. */
void generateOutput(FILE * outfile)
{
  Instruction * instruction;

  /* First reverse the order in which the instructions
     occur (becaused they were pushed in LIFO order). */
  reverseInstructionList();

  /* Reset PC */
  pc = 0L;

  for(instruction = instructionList;
      instruction != NIL;
      instruction = instruction->next)
    emitInstruction(outfile, instruction);
}



/* Reverse the order of the instruction list.
   This must be called before actual code is emitted. */
static void reverseInstructionList(void)
{
  Instruction * instruction, * next, * prev;
  
  prev = NIL;
  for(instruction = instructionList;
      instruction != NIL;
      prev = instruction, instruction = next)
  {
    next = instruction->next;
    instruction->next = prev;
  }
  instructionList = prev;
}


static void emitInstruction(FILE * fileptr, Instruction * instruction)
{
  Argument * argument;
  long value;
  long offset;
  int index;  /* formal argument index */
  int bytes;  /* required bytes (1 or 2) */

  long emitter[3];
  int size[3];
  int args;

  /* Emit the opcode number */
  if(instruction->opcode->opcode >= 0x00)
  {
    emit(fileptr, instruction->opcode->opcode);
    pc++;
  }

  /* Loop through, checking arguments.
     The basic form of the arguments will have been
     checked at the first pass.  This pass just verifies
     that constants are the correct size. */
  index = 0;
  for(argument = instruction->args; argument != NIL;
      argument = argument->next, index++)
  {
    value = evalExpr(argument->expr, instruction->lineNumber);
    if(instruction->opcode->opcode == -3)
    {
      /* DS instruction */
      value = revbits(value);
    }

    /* Check size! */
    switch(instruction->opcode->argtypes[index])
    {
    case REG: 
    case INDIRECT: 
    case CON1:
      bytes = 1; break;
    case CON2:
      bytes = 2; break;
    case RELATIVE:
      bytes = 2; break;
    default:
      error("internal error: bad formal argtype %d",
	    instruction->opcode->argtypes[index]);
    }
    if(value < 0L)
      errorAt(instruction->lineNumber,
	      "negative values are not allowed (%ld)", value);
    else if(bytes == 1)
    {
      if(value > 0xFF)  /* 1 byte */
	errorAt(instruction->lineNumber,
		"Value too large (%ld), max is 255", value);
    }
    else if(bytes == 2)
    {
      if(value > 0xFFFF)  /* 2 bytes */
	errorAt(instruction->lineNumber,
		"Value too large (%ld), max is 65535", value);
    }

    size[index] = bytes;

    /* Emit argument byte(s) based on type: */
    switch(instruction->opcode->argtypes[index])
    {
    case REG:
    case INDIRECT:
    case CON1:
    case CON2:
      emitter[index] = value;
      break;
    case RELATIVE:
      bytes = 1;
      size[index] = bytes;
      offset = (value - pc) - 1; /* startPC) - 2; */
      if(offset < 0)
	offset = 0x100 + offset;
    
      if(offset < 0 || offset > 0xFF)
	errorAt(instruction->lineNumber,
		"relative jump too long");
      emitter[index] = offset;
      break;
    default: break;
    }
    pc += bytes;
  }
  args = index;
  for(index = 0; index < args; index++)
  {
    long val;
    int bytes;
    
    if((index == 0 || index == 1) &&
       instruction->opcode->revargs)
    {
      val = emitter[!index];
      bytes = size[!index];
    }
    else
    {
      val = emitter[index];
      bytes = size[index];
    }

    if(bytes == 1)
      emit(fileptr, val);
    else
    {
      /* LSB first */
      emit(fileptr, val & 0xFF);
      emit(fileptr, (val >> 8) & 0xFF);
    }
  }
}


static void emit(FILE * fileptr, int opcode)
{
  fputc(opcode, fileptr);
}


/* Reverse the first 16 bits in the value */
static long revbits(long val)
{
  long src, dest;
  long index;
  long bitsrc, bitdest;
  
  src = val & 0xFFFF;
  dest = 0;
  val ^= src;
  
  for(index = 0; index < 16; index++)
  {
    bitsrc =  1L << index;
    bitdest = 1L << (15 - index);
    if(src & bitsrc)
      dest |= bitdest;
  }
  val |= dest;
  return (val);
}



/* Evaluate an expression completely.  Gives an error and
   aborts if the expression is not exactly determined. */
static long evalExpr(Expr * expr, int line)
{
  Label * label;
  long val;
  
  switch(expr->type)
  {
  case INTEGER_EXPR:
    return (expr->value.integer);
  case LABEL_EXPR:
    label = findLabel(expr->value.label);
    if(label == NIL)
      errorAt(line, "internal error: label \"%s\" not found",
	    expr->value.label);
    if(label->resolved == FALSE)
      errorAt(line, "unresolved label \"%s\"", expr->value.label);
    return (evalExpr(label->value, line));
  case ADD_EXPR:
    return (evalExpr(expr->value.operands.left, line) +
	    evalExpr(expr->value.operands.right, line));
  case SUBTRACT_EXPR:
    return (evalExpr(expr->value.operands.left, line) -
	    evalExpr(expr->value.operands.right, line));
  case MULTIPLY_EXPR:
    return (evalExpr(expr->value.operands.left, line) *
	    evalExpr(expr->value.operands.right, line));
  case DIVIDE_EXPR:
    val = evalExpr(expr->value.operands.right, line);
    if(val == 0L)
      errorAt(line, "division by zero");
    return (evalExpr(expr->value.operands.left, line) / val);
  case MODULUS_EXPR:
    val = evalExpr(expr->value.operands.right, line);
    if(val == 0L)
      errorAt(line, "modulus by zero");
    return (evalExpr(expr->value.operands.left, line) / val);
  case AND_EXPR:
    return (evalExpr(expr->value.operands.left, line) &
	    evalExpr(expr->value.operands.right, line));
  case OR_EXPR:
    return (evalExpr(expr->value.operands.left, line) |
	    evalExpr(expr->value.operands.right, line));
  case XOR_EXPR:
    return (evalExpr(expr->value.operands.left, line) ^
	    evalExpr(expr->value.operands.right, line));
  case UMINUS_EXPR:
    return (-evalExpr(expr->value.operands.left, line));
  default:
    errorAt(line, "internal error: unrecognized expression type %d",
	    expr->type);
    return (0L);
  }
}



